# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

from django.core.management import BaseCommand, CommandError

from cronman.scheduler import CronScheduler


class Command(BaseCommand):
    help = (
        "Cron Scheduler: determines which CronJobs should be started in "
        "current period and starts each of them in a worker process."
    )

    def add_arguments(self, parser):
        parser.add_argument("method", choices=("run", "disable", "enable"))
        parser.add_argument("--workers", action="store_true")

    def handle(self, **options):
        """Main command logic"""
        method_name = options["method"]
        workers = options["workers"]
        scheduler = CronScheduler()
        method = getattr(scheduler, method_name)
        if method_name == "run":
            if workers:
                raise CommandError(
                    'Subcommand "{}" does not accept --workers option.'.format(
                        method_name
                    )
                )
            return method()
        else:
            return method(workers=workers)



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

import errno
import os
import sys
import time

from cronman.base import BaseCronObject
from cronman.config import app_settings
from cronman.job import cron_job_registry
from cronman.utils import bool_param, config, parse_job_spec, spawn


class CronSpawner(BaseCronObject):
    """Cron Spawner class - responsible for starting new worker processes"""

    wait_for_memory = 7  # number of seconds to wait on first OOM error

    def __init__(self, extra_env=None, **kwargs):
        super(CronSpawner, self).__init__(**kwargs)
        self.extra_env = extra_env or {}
        self.memory_error_occurred = False

    def get_worker_env(self):
        """Constructs a dictionary of environment variables for worker
        subprocess.
        This way we can ensure that cron-specific settings are identical
        for scheduler and workers.
        """

        # Note: We need to ensure that values stored in env are
        # string or bytestring

        environ = os.environ.copy()
        environ["CRONMAN_JOBS_MODULE"] = str(config("CRONMAN_JOBS_MODULE"))
        environ["CRONMAN_DATA_DIR"] = str(self.data_dir)
        environ["CRONMAN_DEBUG"] = str(
            int(bool_param(config("CRONMAN_DEBUG"), default=False))
        )
        environ["CRONMAN_NICE_CMD"] = str(config("CRONMAN_NICE_CMD") or "")
        environ["CRONMAN_IONICE_CMD"] = str(config("CRONMAN_IONICE_CMD") or "")
        environ["CRONMAN_CRONITOR_URL"] = str(config("CRONMAN_CRONITOR_URL"))
        environ["CRONMAN_CRONITOR_ENABLED"] = str(
            int(bool_param(config("CRONMAN_CRONITOR_ENABLED"), default=False))
        )
        environ["CRONMAN_SLACK_ENABLED"] = str(
            int(bool_param(config("CRONMAN_SLACK_ENABLED"), default=False))
        )
        environ["CRONMAN_SENTRY_ENABLED"] = str(
            int(bool_param(config("CRONMAN_SENTRY_ENABLED"), default=False))
        )
        environ.update(self.extra_env)
        return environ

    def get_process_priority_args(self, job_spec):
        """Constructs a list of arguments to be prepended to process args
        in order to assign CPU/IO priority.
        """
        cron_job_class = cron_job_registry.get(parse_job_spec(job_spec)[0])
        if (
            app_settings.CRONMAN_NICE_CMD
            and cron_job_class.worker_cpu_priority is not None
        ):
            cpu_priority_args = [
                app_settings.CRONMAN_NICE_CMD,
                "-n",
                str(cron_job_class.worker_cpu_priority),
            ]
        else:
            cpu_priority_args = []
        if (
            app_settings.CRONMAN_IONICE_CMD
            and cron_job_class.worker_io_priority is not None
        ):
            io_class, io_class_data = cron_job_class.worker_io_priority
            io_priority_args = [
                app_settings.CRONMAN_IONICE_CMD,
                "-c",
                str(io_class),
            ]
            if io_class_data is not None:
                io_priority_args += ["-n", str(io_class_data)]
        else:
            io_priority_args = []
        return cpu_priority_args + io_priority_args

    def start_worker(self, job_spec):
        """Starts a worker process for given job spec"""
        # Building process parameters:
        kwargs = {"env": self.get_worker_env()}
        args = [sys.executable, sys.argv[0], "cron_worker", "run", job_spec]
        options = [a for a in sys.argv if a.startswith("--settings=")]
        if self.sentry.raven_cmd:
            # All worker processes should be executed by raven-cmd:
            args[-1] = args[-1].replace('"', "'")
            args = self.get_process_priority_args(job_spec) + args + options
            args = [('"{}"'.format(a) if " " in a else a) for a in args]
            args = [self.sentry.raven_cmd, "-c", " ".join(args)]
        else:
            args = self.get_process_priority_args(job_spec) + args + options
        # Spawning a new subprocess
        # (with special case for temporary memory error):
        pid = None
        tries = 1 if self.memory_error_occurred else 2
        while tries:
            tries -= 1
            try:
                pid = spawn(*args, **kwargs)
            except OSError as error:
                if error.errno != errno.ENOMEM:
                    raise
                self.memory_error_occurred = True
                if tries:
                    self.logger.debug(
                        "Unable to start a worker process "
                        "for {} due to Out-Of-Memory error."
                        "Retrying in {} seconds...".format(
                            job_spec, self.wait_for_memory
                        )
                    )
                    time.sleep(self.wait_for_memory)
                else:
                    self.warning(
                        RuntimeError(
                            "Unable to start a worker process "
                            "for {} due to Out-Of-Memory error.".format(
                                job_spec
                            )
                        )
                    )
                    break
            else:
                break
        return pid



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

import os
import time

from django.utils.encoding import force_bytes

from cronman.job import BaseCronJob
from cronman.taxonomies import CPUPriority, IOPriority, LockType

# Test cron jobs:


class Sleep(BaseCronJob):
    """Test CronJob: sleeps for given number of seconds.
    No lock - concurrent calls allowed.
    """

    lock_type = None  # concurrency allowed
    can_resume = False

    def run(self, seconds=None, path=None, **kwargs):
        """Main logic"""
        time.sleep(int(seconds) if seconds else 0)
        if path:
            with open(path, "wb") as file_:
                kwargs["seconds"] = seconds or ""
                kwargs["path"] = path or ""
                for key, value in kwargs.items():
                    file_.write(force_bytes("{}={}\n".format(key, value)))
                for key, value in os.environ.items():
                    file_.write(force_bytes("{}={}\n".format(key, value)))
                file_.write(
                    force_bytes("Slept for {} second(s).\n".format(seconds))
                )
                file_.write(b"Done.\n")


class ClassLockedSleep(Sleep):
    """Test CronJob: sleeps for given number of seconds.
    Lock by class - concurrent calls of this class are no allowed.
    """

    lock_type = LockType.CLASS


class ParamsLockedSleep(Sleep):
    """Test CronJob: sleeps for given number of seconds.
    Lock by params - concurrent calls of this class are allowed only if
    they use different arguments.
    """

    lock_type = LockType.PARAMS


class IgnoreLockErrorsSleep(Sleep):
    """Test CronJob: sleeps for given number of seconds.
    Ignore lock errors - report lock errors as warnings.
    Lock by params - concurrent calls of this class are allowed only if
    they use different arguments.
    """

    lock_type = LockType.PARAMS
    lock_ignore_errors = True


class SlackNotifyDoneSleep(Sleep):
    """Test CronJob: sleeps for given number of seconds.
    Slack notification is sent by the worker when job is done.
    """

    slack_notify_done = True


class LowCPUSleep(Sleep):
    """Test CronJob: sleeps for given number of seconds.
    Low CPU priority (`nice -n 10`)
    """

    worker_cpu_priority = CPUPriority.LOW


class LowestCPUIOSleep(Sleep):
    """Test CronJob: sleeps for given number of seconds.
    Lowest CPU and IO priority (`nice -n 19 ionice -c 2 -n 7`)

    NOTE: This is just an example, in real life there is no need to set
    `IOPriority.BEST_EFFORT_LOWEST` when `CPUPriority.LOWEST` is set because
    this IO priority level is implied.
    """

    worker_cpu_priority = CPUPriority.LOWEST
    worker_io_priority = IOPriority.BEST_EFFORT_LOWEST


class IdleIOSleep(Sleep):
    """Test CronJob: sleeps for given number of seconds.
    Idle IO priority.
    """

    worker_io_priority = IOPriority.IDLE


class PersistentSleep(Sleep):
    """Test CronJob: sleeps for given number of seconds.
    Has resume capability.
    """

    lock_type = LockType.CLASS
    can_resume = True


class PersistentSleep2(Sleep):
    """Test CronJob: sleeps for given number of seconds.
    Has resume capability. (2)
    """

    lock_type = LockType.PARAMS
    can_resume = True



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals


class LockType(object):
    """Type of lock acquired by Worker"""

    CLASS = "class"  # one lock per CronJob class
    PARAMS = "params"  # one lock per CronJob class + hash of params


class PIDStatus(object):
    """Status of PID file and associated process"""

    ALIVE = "ALIVE"  # PID file exists and corresponding process is alive
    DEAD = "DEAD"  # PID file exists but there is no corresponding process
    TERMED = "TERMED"  # Killed by SIGTERM
    KILLED = "KILLED"  # Killed by SIGKILL
    DELETED = "DELETED"  # PID file has been deleted


class JobSpecStatus(object):
    """Status of JobSpec file and associated process"""

    STALLED = "STALLED"  # JobSpec file exists and process is dead
    ACTIVE = "ACTIVE"  # JobSpec file exists and process is alive
    DELETED = "DELETED"  # JobSpec file has been deleted
    RESUMED = "RESUMED"  # Process has been started (resumed)


class CronTaskStatus(object):
    """Status of a Cron Task object"""

    WAITING = "waiting"
    QUEUED = "queued"
    STARTED = "started"
    FINISHED = "finished"
    FAILED = "failed"

    CHOICES = (
        (WAITING, "Waiting"),
        (QUEUED, "Queued"),
        (STARTED, "Started"),
        (FINISHED, "Finished"),
        (FAILED, "Failed"),
    )


class CPUPriority(object):
    """CPU Priority of Worker process (`nice`)"""

    LOWEST = 19
    LOW = 10
    NORMAL = 0  # Explicit default value
    # NOTE: higher priority values may require root access to be set
    # HIGH = -10
    # HIGHEST = -20


class IOPriority(object):
    """IO Priority of Worker process (`ionice`)"""

    IDLE = (3, None)  # Receives IO access only when devices are free

    # Best effort - regular IO priority:
    BEST_EFFORT_LOWEST = (2, 7)  # equivalent of `nice -n 19`
    BEST_EFFORT_LOW = (2, 6)  # equivalent of `nice -n 10`
    BEST_EFFORT_NORMAL = (2, 4)  # equivalent of `nice -n 0` (default)
    BEST_EFFORT_HIGH = (2, 2)  # equivalent of `nice -n -10`
    BEST_EFFORT_HIGHEST = (2, 0)  # equivalent of `nice -n -20`

    # Real time - process will ALWAYS receive IO access:
    # NOTE: these options are potentially dangerous
    # REAL_TIME_LOWEST = (1, 7)
    # REAL_TIME_LOW = (1, 6)
    # REAL_TIME_NORMAL = (1, 4)
    # REAL_TIME_HIGH = (1, 2)
    # REAL_TIME_HIGHEST = (1, 0)


class CronSchedulerStatus(object):
    """Cron Scheduler status choices"""

    DISABLED = "disabled"
    ENABLED = "enabled"



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

import datetime
import logging

from django.utils.functional import cached_property

from croniter import croniter

from cronman.base import BaseCronObject
from cronman.config import app_settings
from cronman.exceptions import (
    CronSchedulerLocked,
    CronSchedulerNoJobs,
    CronSchedulerUnlocked,
)
from cronman.monitor import send_errors_to_sentry
from cronman.remote_manager import CronRemoteManager
from cronman.scheduler.files import (
    CronSchedulerLockFile,
    CronSchedulerResumeFile,
)
from cronman.spawner import CronSpawner
from cronman.taxonomies import CronSchedulerStatus
from cronman.utils import cron_jobs_module_config, format_exception
from cronman.worker import CronWorker

logger = logging.getLogger("cronman.command.cron_scheduler")


class CronScheduler(BaseCronObject):
    """Cron Scheduler class.

    Responsible for:
    * determining which jobs should be started in given moment,
    * starting worker processes,
    * maintaining own lock.
    """

    interval = 2  # number of seconds between each scheduler call
    wait_for_memory = 7  # number of seconds to wait on first OOM error

    def __init__(self, now=None, **kwargs):
        self.cronitor_id = app_settings.CRONMAN_CRON_SCHEDULER_CRONITOR_ID
        kwargs["logger"] = kwargs.get("logger", logger)
        super(CronScheduler, self).__init__(**kwargs)
        self.now = now or datetime.datetime.now()
        self.lock_file = CronSchedulerLockFile(self.data_dir)
        self.resume_file = CronSchedulerResumeFile(self.data_dir)
        self.remote_manager = CronRemoteManager()

    @cached_property
    def cron_jobs(self):
        """Return jobs if CRONMAN_JOBS_MODULE is a valid module (must have
        CRON_JOBS attribute). Return () otherwise.
        """
        return cron_jobs_module_config("CRON_JOBS", default=())

    @cached_property
    def cron_spawner(self):
        """Cron Spawner instance"""
        return CronSpawner(
            data_dir=self.data_dir, debug=self.debug, logger=self.logger
        )

    @cached_property
    def cron_worker(self):
        """Cron Worker instance"""
        return CronWorker(
            data_dir=self.data_dir, debug=self.debug, logger=self.logger
        )

    @send_errors_to_sentry
    def run(self):
        """Starts worker processes for jobs that should start in this moment"""

        if not self.cron_jobs:
            self.warning(
                CronSchedulerNoJobs(
                    "Scheduler has no jobs to start. "
                    "Please verify settings.CRONMAN_JOBS_MODULE."
                )
            )

        killed_jobs = self.remote_manager.pop_killed()
        for job_spec_or_pid in killed_jobs:
            self.logger.info(
                'Scheduler: processing KILL "{}" request from '
                "Remote Manager...".format(job_spec_or_pid)
            )
            kill_output = self.cron_worker.kill(job_spec_or_pid)
            self.logger.info(kill_output)

        remote_status = (
            # Retrieve global (per-cluster) status:
            self.remote_manager.get_status("ALL")
            or
            # Retrieve per-host status and remove it from the queue:
            self.remote_manager.pop_status()
        )

        if self.lock_file.exists():
            if remote_status == CronSchedulerStatus.ENABLED:
                self.logger.info(
                    "Scheduler: processing ENABLE request from "
                    "Remote Manager..."
                )
                # Enable the scheduler, create resume file:
                self.enable(workers=True)
            else:
                return self.warning(
                    CronSchedulerLocked(
                        "Scheduler is disabled (lock file exists). "
                        "To enable it again, please run "
                        '"cron_scheduler enable". Quitting now.'
                    )
                )
        else:
            if remote_status == CronSchedulerStatus.DISABLED:
                self.logger.info(
                    "Scheduler: processing DISABLE request from "
                    "Remote Manager..."
                )
                # Disable scheduler, kill running workers, quit:
                return self.disable(workers=True)

        if self.resume_file.exists():
            self.resume_file.delete()
            output = self.resume_workers()
        else:
            output = ""

        run_start = datetime.datetime.now()
        jobs = self.get_jobs()
        num_jobs = len(jobs)
        num_started = 0

        self.before_start()
        try:
            for i, (time_spec, job_spec) in enumerate(jobs, 1):
                self.logger.info(
                    "Starting worker for {} {} ({}/{})".format(
                        time_spec, job_spec, i, num_jobs
                    )
                )
                pid = self.start_worker(job_spec)
                if pid is not None:
                    num_started += 1
            run_end = datetime.datetime.now()
            if num_started:
                output += "Started {} job(s) in {}\n".format(
                    num_started, run_end - run_start
                )
            else:
                output += "No jobs started.\n"
        except Exception as error:
            self.on_error(error)
            raise error
        else:
            self.on_success()
        return output

    @send_errors_to_sentry
    def disable(self, workers=False):
        """Disables the scheduler, so future calls to `run` will not start any
        workers. When `workers` option is set, it also kills all running
        workers via `CronWorker.suspend`.
        """
        if self.lock_file.exists():
            return self.warning(
                CronSchedulerLocked(
                    "Scheduler is already disabled (lock file exists)."
                )
            )
        self.lock_file.create()
        summary = ["lock file created"]

        if workers:
            suspend_output = self.cron_worker.suspend()
            summary.append("workers suspended")
        else:
            suspend_output = ""

        return "Scheduler disabled ({}).\n{}".format(
            ", ".join(summary), suspend_output
        )

    @send_errors_to_sentry
    def enable(self, workers=False):
        """Enables the scheduler, so future calls to `run` may start new
        workers. When `workers` option is set, it also creates a resume file,
        so next call to `run` will also start killed workers via
        `CronWorker.resume`.
        """
        if not self.lock_file.exists():
            return self.warning(
                CronSchedulerUnlocked(
                    "Scheduler is already enabled (lock file does not exist)."
                )
            )
        summary = []

        if workers:
            self.resume_file.create()
            summary.append("resume file created")

        self.lock_file.delete()
        summary.append("lock file deleted")

        return "Scheduler enabled ({}).\n".format(", ".join(summary))

    # Helpers:

    def get_datetime_range(self):
        """Start and end datetime for current scheduler call"""
        start = self.now.replace(second=0, microsecond=0) - datetime.timedelta(
            seconds=1
        )
        end = start + datetime.timedelta(minutes=self.interval)
        return start, end

    def get_jobs(self):
        """List of tuples (time spec, job spec) for jobs that should be started
        in current scheduler call.
        Jobs are sorted by start datetime, earliest first.
        """
        start, end = self.get_datetime_range()
        to_be_started = []
        for time_spec, job_spec in self.cron_jobs:
            job_start = croniter(time_spec, start).get_next(datetime.datetime)
            if job_start <= end:
                to_be_started.append((job_start, time_spec, job_spec))
        return [
            (time_spec, job_spec)
            for _job_start, time_spec, job_spec in sorted(to_be_started)
        ]

    def start_worker(self, job_spec):
        """Starts a worker process for given job spec"""
        return self.cron_spawner.start_worker(job_spec)

    def resume_workers(self):
        """Resumes killed workers"""
        return self.cron_worker.resume()

    # Cron monitoring implementation

    def before_start(self):
        """Method is used to notify Cronitor just before cron scheduler is
        about to start. Works if `CRONMAN_CRON_SCHEDULER_CRONITOR_ID` setting
        is provided, else does nothing.
        """
        if self.cronitor_id:
            self.cronitor.run(self.cronitor_id)

    def on_success(self):
        """Method is used to notify Cronitor just after cron scheduler
        finishes work without exception. Works if
        `CRONMAN_CRON_SCHEDULER_CRONITOR_ID` setting is provided, else does
        nothing.
        """
        if self.cronitor_id:
            self.cronitor.complete(self.cronitor_id)

    def on_error(self, error):
        """Method is used to notify Cronitor just after an exception is caught
        on cron scheduler `run`. Works if `CRONMAN_CRON_SCHEDULER_CRONITOR_ID`
        setting is provided, else does nothing.
        """
        if self.cronitor_id:
            message = format_exception(error)
            self.logger.error(message)
            self.cronitor.fail(self.cronitor_id, msg=message)



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

import errno
import hashlib
import os
import shutil
import signal

from django.test.testcases import TestCase
from django.test.utils import override_settings
from django.utils.encoding import force_bytes

import mock

from cronman.config import app_settings
from cronman.worker import CronWorker

TEMP_FILE = "/tmp/sleep.txt"
TEST_CRONMAN_DATA_DIR = os.path.join(app_settings.CRONMAN_DATA_DIR, "test")


def override_cron_settings(**kwargs):
    """Override settings with default values for `cronman` app"""
    defaults = {
        "CRONMAN_DATA_DIR": TEST_CRONMAN_DATA_DIR,
        "CRONMAN_DEBUG": True,
        "CRONMAN_JOBS_MODULE": "cronman.tests.cron_jobs",
        "CRONMAN_CRONITOR_ENABLED": False,
        "CRONMAN_SLACK_ENABLED": False,
        "CRONMAN_SENTRY_ENABLED": True,
        "CRONMAN_RAVEN_CMD": None,
        "CRONMAN_NICE_CMD": "nice",
        "CRONMAN_IONICE_CMD": "ionice",
        "CRONMAN_REMOTE_MANAGER_ENABLED": False,
    }
    defaults.update(kwargs)
    return override_settings(**defaults)


def create_pid_file(job_spec, pid=None):
    """Utility function to create PID file for given job spec"""
    pid = os.getpid() if pid is None else pid
    worker = CronWorker()
    name, args, kwargs, cron_job_class = worker.parse_job_spec_with_class(
        job_spec
    )
    pid_file = worker.get_pid_file(cron_job_class, name, args, kwargs)
    with mock.patch("cronman.worker.worker_file.os.getpid", lambda: pid):
        pid_file.create()
    return pid_file


def create_job_spec_file(job_spec, content=None):
    """Utility function to create JobSpec file for given job spec"""
    content = job_spec if content is None else content
    worker = CronWorker()
    name, args, kwargs, cron_job_class = worker.parse_job_spec_with_class(
        job_spec
    )
    pid_file = worker.get_pid_file(cron_job_class, name, args, kwargs)
    job_spec_file = pid_file.job_spec_file
    job_spec_file.create(content)
    return job_spec_file


def get_params_hash(args, kwargs):
    """Generate shortened hash of cron job parameters."""
    params_bytes_repr = force_bytes(repr((args, kwargs)))
    return hashlib.md5(params_bytes_repr).hexdigest()[:10]


def patch_ps(active_pids=(), zombie_pids=()):
    """Patches `ps` calls made by ProcessManager"""

    def mock_check_output(*args, **kwargs):
        pid = int(args[0][2])
        if pid in active_pids:
            status = "R"
        elif pid in zombie_pids:
            status = "Z"
        else:
            status = ""
        return "{}\n".format(status)

    return mock.patch(
        "cronman.worker.process_manager.subprocess.check_output",
        mock_check_output,
    )


def patch_kill(active_pids=(), die_on_sigterm_pids=(), die_on_sigkill_pids=()):
    """Patches `os.kill` calls made by ProcessManager"""

    active_pids = list(active_pids)  # this have to be mutable

    def mock_kill(pid, sig):
        if pid not in active_pids:
            raise OSError(errno.ESRCH, "No such process")
        if sig == signal.SIGTERM and pid in die_on_sigterm_pids:
            active_pids.remove(pid)
        if sig == signal.SIGKILL and pid in die_on_sigkill_pids:
            active_pids.remove(pid)

    return mock.patch("cronman.worker.process_manager.os.kill", mock_kill)


def mock_environ():
    """Mock for `os.environ.copy`"""
    return {"SOME_ENV_VAR": "42"}


def expected_worker_env():
    """Environment dictionary expected to be passed to worker subprocess"""
    cronitor_url = "https://cronitor.link/{cronitor_id}/{end_point}"
    return {
        "SOME_ENV_VAR": "42",
        "CRONMAN_CRONITOR_ENABLED": "0",
        "CRONMAN_CRONITOR_URL": cronitor_url,
        "CRONMAN_JOBS_MODULE": "cronman.tests.cron_jobs",
        "CRONMAN_DATA_DIR": TEST_CRONMAN_DATA_DIR,
        "CRONMAN_DEBUG": "1",
        "CRONMAN_SLACK_ENABLED": "0",
        "CRONMAN_NICE_CMD": "nice",
        "CRONMAN_IONICE_CMD": "ionice",
        "CRONMAN_SENTRY_ENABLED": "1",
    }


class BaseCronTestCase(TestCase):
    """Base class for `cron_*` commands test cases"""

    def setUp(self):
        super(BaseCronTestCase, self).setUp()
        self._remove_temp_file()
        self._remove_test_cron_data_dir()

    def tearDown(self):
        self._remove_temp_file()
        self._remove_test_cron_data_dir()
        super(BaseCronTestCase, self).tearDown()

    def _remove_temp_file(self):
        if os.path.exists(TEMP_FILE):
            os.unlink(TEMP_FILE)

    def _remove_test_cron_data_dir(self):
        if os.path.exists(TEST_CRONMAN_DATA_DIR):
            shutil.rmtree(TEST_CRONMAN_DATA_DIR)

    def assertMessageInTempFile(self, message):
        """Check that the message is present in TEMP_FILE"""
        self.assertTrue(os.path.exists(TEMP_FILE))
        with open(TEMP_FILE, "r") as file_:
            lines = [line.strip() for line in file_.readlines()]
        self.assertIn(message, lines)



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from cronman.scheduler.files import CronSchedulerLockFile
from cronman.scheduler.scheduler import CronScheduler

__all__ = ["CronScheduler", "CronSchedulerLockFile"]



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

from django import forms
from django.utils.functional import cached_property

from six import text_type

from cronman.job import cron_job_registry
from cronman.models import CronTask
from cronman.utils import cron_jobs_module_config, parse_params


class CronTaskAdminForm(forms.ModelForm):
    """Admin form for CronTask model"""

    class Meta:
        model = CronTask
        fields = ("cron_job", "params", "start_at")

    cron_job = forms.ChoiceField(choices=())

    def __init__(self, *args, **kwargs):
        super(CronTaskAdminForm, self).__init__(*args, **kwargs)
        if "cron_job" in self.fields:
            self.fields["cron_job"].choices = sorted(
                (name, name)
                for name, cron_job_class in cron_job_registry.items()
                if self.accept_cron_job_choice(name, cron_job_class)
            )

    @cached_property
    def allowed_tasks(self):
        """Allowed task (cron job) names"""
        return cron_jobs_module_config("ALLOWED_CRON_TASKS", default=())

    def clean_params(self):
        """Validate `params` field value"""
        params = self.cleaned_data.get("params") or ""
        try:
            parse_params(params)
        except ValueError as e:
            raise forms.ValidationError(text_type(e))
        return params

    def accept_cron_job_choice(self, name, cron_job_class):
        """Check if given cron job name and class can be accepted as
        `cron_job` choice in this form.
        """
        return name in self.allowed_tasks



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

import logging
import re
import sys
import time
import traceback

from django.utils import timezone

from cronman.base import BaseCronObject
from cronman.exceptions import (
    CronJobNotRegistered,
    CronTaskInvalidStatus,
    CronWorkerInvalidParams,
    CronWorkerLocked,
)
from cronman.job import cron_job_registry
from cronman.models import CronTask
from cronman.monitor import send_errors_to_sentry
from cronman.taxonomies import LockType
from cronman.utils import TabularFormatter, format_exception, parse_job_spec
from cronman.worker.cron_job_info import CronJobClassList
from cronman.worker.process_manager import ProcessManager
from cronman.worker.signal_notifier import SignalNotifier
from cronman.worker.worker_file import CronWorkerPIDFile
from cronman.worker.worker_list import CronWorkerJobSpecList, CronWorkerPIDList

logger = logging.getLogger("cronman.command.cron_worker")


class CronWorker(BaseCronObject):
    """Cron Worker class.

    Responsible for:
    * running a CronJob class defined by job spec,
    * passing status to external logging services,
    * maintaining pidfile / lock,
    * listing active worker processes,
    * killing processes.
    """

    NO_PID_FILES_MESSAGE = "No PID file(s) found."
    NO_JOB_SPEC_FILES_MESSAGE = "No JobSpec file(s) found."
    NO_CRON_JOBS_MESSAGE = "No cron job(s) found."

    def __init__(self, **kwargs):
        self.cronitor_id = None
        self.formatter = TabularFormatter()
        kwargs["logger"] = kwargs.get("logger", logger)
        super(CronWorker, self).__init__(**kwargs)

    @staticmethod
    def get_cronitor_id(kwargs):
        """Retrieves Cronitor ID passed through cron job params,
        removes it from the params.
        """
        return (kwargs or {}).pop("cronitor_id", None)

    @send_errors_to_sentry
    def run(self, job_spec):
        """Executes logic defined in CronJob class specified by `job_spec`"""
        name, args, kwargs, cron_job_class = self.parse_job_spec_with_class(
            job_spec
        )
        self.cronitor_id = self.get_cronitor_id(kwargs)
        cron_task = self.get_cron_task(kwargs)
        if cron_task and not cron_task.is_pending():
            if self.cron_task_killed(cron_task):
                self.logger.info(
                    'Starting "{}" for killed CronTask.'.format(job_spec)
                )
            else:
                error = CronTaskInvalidStatus(
                    'Unable to start "{}", because associated CronTask '
                    'has invalid status "{}".'.format(
                        job_spec, cron_task.get_status_display()
                    )
                )
                return self.warning(error)

        if cron_task:
            cron_task.mark_as_queued()

        pid_file = self.get_pid_file(cron_job_class, name, args, kwargs)
        if self.pid_file_locked(pid_file, cron_job_class.lock_check_attempts):
            error = CronWorkerLocked(
                'Unable to start "{}", because similar process '
                "is already running (PID file exists).".format(job_spec)
            )
            return self.warning(
                error, silent=cron_job_class.lock_ignore_errors
            )
        job_spec_file = self.get_job_spec_file(cron_job_class, pid_file)

        with SignalNotifier(job_spec):

            pid_file.create()
            if job_spec_file:
                job_spec_file.create(job_spec)

            run_start = timezone.now()
            if cron_task:
                cron_task.mark_as_started(pid_file.pid, run_start)
            self.logger.info('Starting "{}"...'.format(job_spec))

            ok = self.run_cron_job(job_spec, cron_job_class, args, kwargs)

            run_end = timezone.now()
            duration = run_end - run_start
            if ok:
                if cron_task:
                    cron_task.mark_as_finished(run_end)
                self.logger.info(
                    'Processing "{}" finished after {}'.format(
                        job_spec, duration
                    )
                )
            else:
                if cron_task:
                    cron_task.mark_as_failed()
                self.logger.warning(
                    'Processing "{}" FAILED after {}'.format(
                        job_spec, duration
                    )
                )

            if job_spec_file:
                job_spec_file.delete()
            pid_file.delete()

        return "{}: Processed {}".format("OK" if ok else "FAIL", job_spec)

    @send_errors_to_sentry
    def status(self, job_spec_or_pid=None):
        """Shows status of all running worker processes,
        optionally filtered by job_spec or PID
        """
        worker_pid_list = self.get_worker_pid_list(job_spec_or_pid)
        items, totals = worker_pid_list.status()
        return self.formatter.format_listing_output(
            items,
            totals=totals,
            title="STATUS:",
            empty_message=self.NO_PID_FILES_MESSAGE,
        )

    def _kill(self, job_spec_or_pid=None):
        """Kills all worker processes,
        optionally filtered by job_spec or PID
        """
        worker_pid_list = self.get_worker_pid_list(job_spec_or_pid)
        items, totals = worker_pid_list.kill()
        return self.formatter.format_listing_output(
            items,
            totals=totals,
            title="KILL:",
            empty_message=self.NO_PID_FILES_MESSAGE,
        )

    kill = send_errors_to_sentry(_kill)

    def _clean_pid_files(self):
        """Removes all dead PID files"""
        items, totals = self.get_worker_pid_list().clean()
        return self.formatter.format_listing_output(
            items,
            totals=totals,
            title="CLEAN PID FILES:",
            empty_message=self.NO_PID_FILES_MESSAGE,
        )

    def _clean_job_spec_files(self):
        """Removes all stalled JobSpec files"""
        items, totals = self.get_worker_job_spec_list().clean()
        return self.formatter.format_listing_output(
            items,
            totals=totals,
            title="CLEAN JOBSPEC FILES:",
            empty_message=self.NO_JOB_SPEC_FILES_MESSAGE,
        )

    def _clean(self):
        """Removes all dead PID files and stalled JobSpec files"""
        return self._clean_pid_files() + self._clean_job_spec_files()

    clean = send_errors_to_sentry(_clean)

    @send_errors_to_sentry
    def suspend(self):
        """Shortcut command to get:
        1. `clean` - remove all DEAD PID files and STALLED JobSpec files
        2. `kill` - kill ALL running worker processes.
        Returns joined output of called commands.
        """
        return self._clean() + self._kill()

    @send_errors_to_sentry
    def resume(self, job_spec_or_pid=None):
        """Starts all previously killed worker processes with `can_resume`
        capability,
        optionally filtered by job_spec or PID
        """
        worker_job_spec_list = self.get_worker_job_spec_list(job_spec_or_pid)
        items, totals = worker_job_spec_list.resume()
        return self.formatter.format_listing_output(
            items,
            totals=totals,
            title="RESUME:",
            empty_message=self.NO_JOB_SPEC_FILES_MESSAGE,
        )

    @send_errors_to_sentry
    def info(self, name=None):
        """Shows a list of all available cron job class,
        or if name is provided, show its parameters and docstring
        """
        name, cron_job_class = self.parse_job_spec_with_class(name)[::3]
        cron_job_list = CronJobClassList(
            name=name, cron_job_class=cron_job_class
        )
        if name:
            items, totals = cron_job_list.details()[0], None
            vertical = True
        else:
            items, totals = cron_job_list.summary()
            vertical = False
        return self.formatter.format_listing_output(
            items,
            totals=totals,
            vertical=vertical,
            empty_message=self.NO_CRON_JOBS_MESSAGE,
        )

    # Cron Job running internals:

    def run_cron_job(self, job_spec, cron_job_class, args, kwargs):
        """Initializes and runs a CronJob"""
        self.before_start(job_spec, cron_job_class, args, kwargs)
        try:
            results = cron_job_class().run(*args, **kwargs)
        except Exception as e:
            self.on_error(job_spec, cron_job_class, args, kwargs, e)
            ok = False
        else:
            self.on_success(job_spec, cron_job_class, args, kwargs, results)
            ok = True
        return ok

    def before_start(self, job_spec, cron_job_class, args, kwargs):
        """Notifies monitor services when job is about to start"""
        cronitor_id = self.cronitor_id or cron_job_class.cronitor_id
        if cronitor_id and cron_job_class.cronitor_ping_run:
            self.cronitor.run(cronitor_id)

    def on_success(self, job_spec, cron_job_class, args, kwargs, result):
        """Notifies monitor services when job succeeded"""
        cronitor_id = self.cronitor_id or cron_job_class.cronitor_id
        if cronitor_id:
            self.cronitor.complete(cronitor_id)
        if cron_job_class.slack_notify_done:
            self.slack.post('Cron job "{}" is done.'.format(job_spec))

    def on_error(self, job_spec, cron_job_class, args, kwargs, error):
        """Notifies monitor services when job failed"""
        cronitor_id = self.cronitor_id or cron_job_class.cronitor_id
        message = format_exception(error)
        self.logger.error(message)
        self.sentry.capture_exception(error)
        if cronitor_id and cron_job_class.cronitor_ping_fail:
            self.cronitor.fail(cronitor_id, msg=message)
        if self.debug:
            traceback.print_exc(file=sys.stderr)

    # Helpers:

    @staticmethod
    def get_cron_task(kwargs):
        """Retrieves CronTask by ID passed through cron job params.
        Removes CronTask ID from params.
        """
        cron_task_id = (kwargs or {}).pop(CronTask.TASK_ID_PARAM, None)
        if cron_task_id:
            cron_task = CronTask.objects.filter(pk=cron_task_id).first()
        else:
            cron_task = None
        return cron_task

    @staticmethod
    def cron_task_killed(cron_task):
        """Checks if given Cron Task is in killed state
        (status STARTED + inactive PID).
        """
        return (
            cron_task.is_started()
            and cron_task.pid
            and not ProcessManager(cron_task.pid).alive()
        )

    def get_pid_file(self, cron_job_class, name, args, kwargs):
        """Retrieves PID file for given CronJob and its parameters"""
        name = cron_job_class.lock_name or name
        if cron_job_class.lock_type == LockType.CLASS:
            pid_file_name = CronWorkerPIDFile.get_file_name(name)
        elif cron_job_class.lock_type == LockType.PARAMS:
            pid_file_name = CronWorkerPIDFile.get_file_name(name, args, kwargs)
        else:  # No lock, just PID file
            pid_file_name = CronWorkerPIDFile.get_file_name(
                name, args, kwargs, random=True
            )
        return CronWorkerPIDFile(self.data_dir, pid_file_name)

    def get_job_spec_file(self, cron_job_class, pid_file):
        """Retrieves JobSpec file for given CronJob class and PIDFile"""
        return pid_file.job_spec_file if cron_job_class.can_resume else None

    @staticmethod
    def pid_file_locked(pid_file, lock_check_attempts):
        """Checks if given PIDFile exists and belongs to running process"""
        locked = True
        while lock_check_attempts > 0:
            lock_check_attempts -= 1
            locked = pid_file.exists_with_alive_process()
            if not locked:
                break
            elif lock_check_attempts:
                time.sleep(1)
        return locked

    def get_worker_pid_list(self, job_spec_or_pid=None):
        """Retrieves CronWorkerPIDList instance"""
        name, args, kwargs, pid = self.parse_job_spec_or_pid(job_spec_or_pid)
        return CronWorkerPIDList(
            self.data_dir, name=name, args=args, kwargs=kwargs, pid=pid
        )

    def get_worker_job_spec_list(self, job_spec_or_pid=None):
        """Retrieves CronWorkerJobSpecList instance"""
        name, args, kwargs, pid = self.parse_job_spec_or_pid(job_spec_or_pid)
        return CronWorkerJobSpecList(
            self.data_dir, name=name, args=args, kwargs=kwargs, pid=pid
        )

    @staticmethod
    def parse_job_spec_with_class(job_spec=None):
        """Converts a string (job spec) into quadruple
        (name, args, kwargs, cron_job_class).
        Raises CronWorkerInvalidParams if string cannot be parsed or
        there is no cron job class for given specification.
        """
        if job_spec:
            try:
                name, args, kwargs = parse_job_spec(job_spec)
                cron_job_class = cron_job_registry.get(name)
            except (ValueError, CronJobNotRegistered) as error:
                raise CronWorkerInvalidParams(format_exception(error))
        else:
            name = args = kwargs = cron_job_class = None
        return name, args, kwargs, cron_job_class

    @staticmethod
    def parse_job_spec_or_pid(job_spec_or_pid=None):
        """Converts a string (job spec or integer - PID) into quadruple
        (name, args, kwargs, pid).
        Raises CronWorkerInvalidParams if string cannot be parsed.
        """
        if job_spec_or_pid:
            if re.match(r"\d+", job_spec_or_pid):
                pid = int(job_spec_or_pid)
                name = args = kwargs = None
            else:
                pid = None
                try:
                    name, args, kwargs = parse_job_spec(job_spec_or_pid)
                except ValueError as error:
                    raise CronWorkerInvalidParams(format_exception(error))
        else:
            name = args = kwargs = pid = None
        return name, args, kwargs, pid



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

import os


class BaseCronSchedulerFile(object):
    """Base class for Cron Scheduler file wrappers"""

    name = None  # must be set in subclass

    def __init__(self, data_dir):
        self.path = os.path.join(data_dir, self.name)

    def create(self):
        """Creates the lock file"""
        open(self.path, "w").close()

    def delete(self):
        """Deletes the lock file"""
        os.unlink(self.path)

    def exists(self):
        """Checks if the lock file exists"""
        return os.path.exists(self.path)


class CronSchedulerLockFile(BaseCronSchedulerFile):
    """Lock file for Cron Scheduler"""

    name = "scheduler.lock"


class CronSchedulerResumeFile(BaseCronSchedulerFile):
    """Resume file for Cron Scheduler"""

    name = "scheduler.resume"



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

import logging

from cronman.exceptions import CronJobAlreadyRegistered, CronJobNotRegistered
from cronman.taxonomies import LockType


class CronJobRegistry(object):
    """Collection of all CronJob classes referenced by name"""

    def __init__(self):
        self._registry = {}

    def register(self, cron_job_class, name=None):
        """Adds a new CronJob class to the registry"""
        name = name or cron_job_class.__name__
        if name in self._registry:
            raise CronJobAlreadyRegistered(name)
        self._registry[name] = cron_job_class

    def unregister(self, name=None, cron_job_class=None):
        """Removes a CronJob class from the registry"""
        name = name or cron_job_class.__name__
        if name not in self._registry:
            raise CronJobNotRegistered(name)
        del self._registry[name]

    def get(self, name):
        """Retrieves a CronJob class from the registry"""
        if name not in self._registry:
            raise CronJobNotRegistered(name)
        return self._registry[name]

    def items(self):
        """Retrieves all registered items"""
        return self._registry.items()


cron_job_registry = CronJobRegistry()


class BaseCronJob(object):
    """Base class for CronJobs"""

    lock_type = LockType.CLASS
    lock_name = None  # May be used to override job name for lock
    lock_check_attempts = 1  # Number of lock check attempts
    lock_ignore_errors = False  # Should we consider lock errors as warnings?
    cronitor_id = None
    cronitor_ping_run = True  # Should we ping Cronitor when job started?
    cronitor_ping_fail = True  # Should we ping Cronitor when job failed?
    slack_notify_done = False  # Post to Slack when job's done?
    worker_cpu_priority = None  # CPU priority for worker processes
    worker_io_priority = None  # IO priority for worker processes
    can_resume = True  # Can we resume this job after suspension?

    def __init__(self, logger=None):
        self.logger = logger or logging.getLogger(
            "cronman.command.{}".format(self.__module__.split(".")[-1])
        )

    def run(self, *args, **kwargs):
        """Main logic"""
        raise NotImplementedError()  # Put your logic here



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

from django.core.management import BaseCommand, CommandError

from cronman.worker import CronWorker


class Command(BaseCommand):
    help = "Cron Worker: runs a single CronJob"

    def add_arguments(self, parser):
        parser.add_argument(
            "method",
            choices=(
                "run",
                "status",
                "kill",
                "info",
                "clean",
                "suspend",
                "resume",
            ),
        )
        parser.add_argument("arg", nargs="?")

    def handle(self, **options):
        """Main command logic"""
        method_name = options["method"]
        arg = options["arg"]

        worker = CronWorker()
        method = getattr(worker, method_name)

        if not arg:
            if method_name == "run":
                raise CommandError("Job specification is required.")
        if arg:
            if method_name in ("clean", "suspend"):
                raise CommandError(
                    'Subcommand "{}" does not accept arguments.'.format(
                        method_name
                    )
                )
            return method(arg)
        else:
            return method()



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

import collections
import re

from django.core.management import call_command

RESULT_PATTERN = re.compile(
    r"(?P<status>[A-Z]+): Processed (?P<job_spec>[\s\S]*)"
)
EXCEPTION_PATTERN = re.compile(
    r"(?P<exc_class_name>[\w\d+]+): (?P<exc_message>[\s\S]*)"
)


CronWorkerRunResult = collections.namedtuple(
    "CronWorkerRunResult",
    ["output", "status", "exc_class_name", "exc_message", "ok"],
)


def call_worker(job_spec):
    """Calls command `cron_worker run <job_spec>` and parses the output"""
    output = call_command("cron_worker", "run", job_spec)
    status = exc_class_name = exc_message = None
    if output:
        result_match = RESULT_PATTERN.match(output)
        if result_match:
            status = result_match.group("status")
        else:
            exc_match = EXCEPTION_PATTERN.match(output)
            if exc_match:
                exc_class_name = exc_match.group("exc_class_name")
                exc_message = exc_match.group("exc_message")
    ok = status == "OK"
    return CronWorkerRunResult(output, status, exc_class_name, exc_message, ok)



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

from django.utils.encoding import force_bytes

import mock
import requests

from cronman.monitor import Cronitor, Slack
from cronman.tests.base import BaseCronTestCase, override_cron_settings


class CronitorTestCase(BaseCronTestCase):
    """Tests for Cronitor class"""

    @override_cron_settings(
        CRONMAN_CRONITOR_URL="https://cronitor.link/{cronitor_id}/{end_point}",
        CRONMAN_CRONITOR_ENABLED=False,
    )
    @mock.patch("cronman.monitor.requests.head")
    def test_run_disabled(self, mock_head):
        """Test for `run` method, case: Cronitor disabled"""
        cronitor = Cronitor()
        cronitor.logger = mock.MagicMock()
        cronitor.run("cRoNiD")
        mock_head.assert_not_called()
        cronitor.logger.warning.assert_has_calls(
            [mock.call("Cronitor request ignored (disabled in settings).")]
        )

    @override_cron_settings(
        CRONMAN_CRONITOR_URL="https://cronitor.link/{cronitor_id}/{end_point}",
        CRONMAN_CRONITOR_ENABLED=True,
    )
    @mock.patch("cronman.monitor.requests.head")
    def test_run_enabled(self, mock_head):
        """Test for `run` method, case: Cronitor enabled"""
        cronitor = Cronitor()
        cronitor.logger = mock.MagicMock()
        cronitor.run("cRoNiD")
        mock_head.assert_called_once_with(
            "https://cronitor.link/cRoNiD/run", params=None, timeout=10
        )

    @override_cron_settings(
        CRONMAN_CRONITOR_URL="https://cronitor.link/{cronitor_id}/{end_point}",
        CRONMAN_CRONITOR_ENABLED=True,
    )
    @mock.patch(
        "cronman.monitor.requests.head",
        side_effect=requests.ConnectTimeout("msg"),
    )
    def test_run_failed(self, mock_head):
        """Test for `run` method, case: Cronitor enabled, request failed"""
        cronitor = Cronitor()
        cronitor.logger = mock.MagicMock()
        cronitor.run("cRoNiD")
        mock_head.assert_called_once_with(
            "https://cronitor.link/cRoNiD/run", params=None, timeout=10
        )
        cronitor.logger.warning.assert_has_calls(
            [
                mock.call(
                    "Cronitor request failed: "
                    "https://cronitor.link/cRoNiD/run "
                    "ConnectTimeout: msg"
                )
            ]
        )

    @override_cron_settings(
        CRONMAN_CRONITOR_URL="https://cronitor.link/{cronitor_id}/{end_point}",
        CRONMAN_CRONITOR_ENABLED=False,
    )
    @mock.patch("cronman.monitor.requests.head")
    def test_complete_disabled(self, mock_head):
        """Test for `complete` method, case: Cronitor disabled"""
        cronitor = Cronitor()
        cronitor.logger = mock.MagicMock()
        cronitor.complete("cRoNiD")
        mock_head.assert_not_called()
        cronitor.logger.warning.assert_has_calls(
            [mock.call("Cronitor request ignored (disabled in settings).")]
        )

    @override_cron_settings(
        CRONMAN_CRONITOR_URL="https://cronitor.link/{cronitor_id}/{end_point}",
        CRONMAN_CRONITOR_ENABLED=True,
    )
    @mock.patch("cronman.monitor.requests.head")
    def test_complete_enabled(self, mock_head):
        """Test for `complete` method, case: Cronitor enabled"""
        cronitor = Cronitor()
        cronitor.logger = mock.MagicMock()
        cronitor.complete("cRoNiD")
        mock_head.assert_called_once_with(
            "https://cronitor.link/cRoNiD/complete", params=None, timeout=10
        )

    @override_cron_settings(
        CRONMAN_CRONITOR_URL="https://cronitor.link/{cronitor_id}/{end_point}",
        CRONMAN_CRONITOR_ENABLED=True,
    )
    @mock.patch(
        "cronman.monitor.requests.head",
        side_effect=requests.ConnectTimeout("msg"),
    )
    def test_complete_failed(self, mock_head):
        """Test for `complete` method, case: Cronitor enabled, request failed"""
        cronitor = Cronitor()
        cronitor.logger = mock.MagicMock()
        cronitor.complete("cRoNiD")
        mock_head.assert_called_once_with(
            "https://cronitor.link/cRoNiD/complete", params=None, timeout=10
        )
        cronitor.logger.warning.assert_has_calls(
            [
                mock.call(
                    "Cronitor request failed: "
                    "https://cronitor.link/cRoNiD/complete "
                    "ConnectTimeout: msg"
                )
            ]
        )

    @override_cron_settings(
        CRONMAN_CRONITOR_URL="https://cronitor.link/{cronitor_id}/{end_point}",
        CRONMAN_CRONITOR_ENABLED=False,
    )
    @mock.patch("cronman.monitor.requests.head")
    def test_fail_disabled(self, mock_head):
        """Test for `fail` method, case: Cronitor disabled"""
        cronitor = Cronitor()
        cronitor.logger = mock.MagicMock()
        cronitor.fail("cRoNiD", msg="RuntimeError: test message")
        mock_head.assert_not_called()
        cronitor.logger.warning.assert_has_calls(
            [mock.call("Cronitor request ignored (disabled in settings).")]
        )

    @override_cron_settings(
        CRONMAN_CRONITOR_URL="https://cronitor.link/{cronitor_id}/{end_point}",
        CRONMAN_CRONITOR_ENABLED=True,
    )
    @mock.patch("cronman.monitor.requests.head")
    def test_fail_enabled(self, mock_head):
        """Test for `fail` method, case: Cronitor enabled"""
        cronitor = Cronitor()
        cronitor.logger = mock.MagicMock()
        cronitor.fail("cRoNiD", msg="RuntimeError: test message")
        mock_head.assert_called_once_with(
            "https://cronitor.link/cRoNiD/fail",
            params={"msg": "RuntimeError: test message"},
            timeout=10,
        )

    @override_cron_settings(
        CRONMAN_CRONITOR_URL="https://cronitor.link/{cronitor_id}/{end_point}",
        CRONMAN_CRONITOR_ENABLED=True,
    )
    @mock.patch(
        "cronman.monitor.requests.head",
        side_effect=requests.ConnectTimeout("msg"),
    )
    def test_fail_failed(self, mock_head):
        """Test for `fail` method, case: Cronitor enabled, request failed"""
        cronitor = Cronitor()
        cronitor.logger = mock.MagicMock()
        cronitor.fail("cRoNiD", msg="RuntimeError: test message")
        mock_head.assert_called_once_with(
            "https://cronitor.link/cRoNiD/fail",
            params={"msg": "RuntimeError: test message"},
            timeout=10,
        )
        cronitor.logger.warning.assert_has_calls(
            [
                mock.call(
                    "Cronitor request failed: "
                    "https://cronitor.link/cRoNiD/fail "
                    "ConnectTimeout: msg"
                )
            ]
        )


class SlackTestCase(BaseCronTestCase):
    """Tests for Slack class"""

    @override_cron_settings(
        CRONMAN_SLACK_URL="https://fake-chat.slack.com/services/hooks/slackbot",
        CRONMAN_SLACK_TOKEN="sLaCkTokEn",
        CRONMAN_SLACK_DEFAULT_CHANNEL="cronitor",
        CRONMAN_SLACK_ENABLED=False,
    )
    @mock.patch("cronman.monitor.requests.post")
    def test_post_disabled(self, mock_post):
        """Test for `post` method, case: Slack disabled"""
        slack = Slack()
        slack.logger = mock.MagicMock()
        slack.post("This is a test!")
        mock_post.assert_not_called()
        slack.logger.warning.assert_has_calls(
            [mock.call("Slack request ignored (disabled in settings).")]
        )

    @override_cron_settings(
        CRONMAN_SLACK_URL="https://fake-chat.slack.com/services/hooks/slackbot",
        CRONMAN_SLACK_TOKEN="sLaCkTokEn",
        CRONMAN_SLACK_DEFAULT_CHANNEL="cronitor",
        CRONMAN_SLACK_ENABLED=True,
    )
    @mock.patch("cronman.monitor.requests.post")
    def test_post_enabled(self, mock_post):
        """Test for `post` method, case: Slack enabled"""
        slack = Slack()
        slack.logger = mock.MagicMock()
        slack.post("This is a test!")
        mock_post.assert_called_once_with(
            "https://fake-chat.slack.com/services/hooks/slackbot?"
            "token=sLaCkTokEn&channel=%23cronitor",
            data=force_bytes("This is a test!"),
            timeout=7,
        )

    @override_cron_settings(
        CRONMAN_SLACK_URL="https://fake-chat.slack.com/services/hooks/slackbot",
        CRONMAN_SLACK_TOKEN="sLaCkTokEn",
        CRONMAN_SLACK_DEFAULT_CHANNEL="cronitor",
        CRONMAN_SLACK_ENABLED=True,
    )
    @mock.patch("cronman.monitor.requests.post")
    def test_post_enabled_custom_channel(self, mock_post):
        """Test for `post` method, case: Slack enabled, custom channel"""
        slack = Slack()
        slack.logger = mock.MagicMock()
        slack.post("This is a test!", channel="dev")
        mock_post.assert_called_once_with(
            "https://fake-chat.slack.com/services/hooks/slackbot?"
            "token=sLaCkTokEn&channel=%23dev",
            data=force_bytes("This is a test!"),
            timeout=7,
        )

    @override_cron_settings(
        CRONMAN_SLACK_URL="https://fake-chat.slack.com/services/hooks/slackbot",
        CRONMAN_SLACK_TOKEN="sLaCkTokEn",
        CRONMAN_SLACK_DEFAULT_CHANNEL="cronitor",
        CRONMAN_SLACK_ENABLED=True,
    )
    @mock.patch(
        "cronman.monitor.requests.post",
        side_effect=requests.ConnectTimeout("msg"),
    )
    def test_post_failed(self, mock_post):
        """Test for `post` method, case: Slack enabled, request failed"""
        slack = Slack()
        slack.logger = mock.MagicMock()
        slack.post("This is a test!")
        mock_post.assert_called_once_with(
            "https://fake-chat.slack.com/services/hooks/slackbot?"
            "token=sLaCkTokEn&channel=%23cronitor",
            data=force_bytes("This is a test!"),
            timeout=7,
        )
        slack.logger.error.assert_has_calls(
            [mock.call("Slack request failed: ConnectTimeout: msg")]
        )



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

import errno
import functools
import logging
import os
import signal
import subprocess

from django.utils.encoding import force_text

logger = logging.getLogger("cronman.command.cron_worker")


def pid_required(otherwise):
    """Decorator for ProcessManager methods to provide alternative result
    when PID is not available.
    """

    def decorator(method):
        @functools.wraps(method)
        def wrapper(self, *args, **kwargs):
            if self.pid is None:
                result = otherwise
            else:
                result = method(self, *args, **kwargs)
            return result

        return wrapper

    return decorator


class ProcessManager(object):
    """Wrapper over PID value, communicates with other processes through
    signals.
    """

    def __init__(self, pid):
        self.pid = pid
        self.logger = logger

    @pid_required(otherwise=False)  # Same as `errno.ESRCH`
    def _kill(self, sig):
        """Sends signal to the process"""
        try:
            os.kill(self.pid, sig)
        except OSError as e:
            if e.errno == errno.ESRCH:
                result = False
            else:  # Process exists but access is denied
                result = None
        else:
            result = True
        return result

    def exists(self):
        """Check if PID is assigned to existing process"""
        return self._kill(0)

    def alive(self):
        """Check if PID is assigned to existing ALIVE process"""
        process_exists = self.exists()
        if process_exists:
            if "Z" in self.status():
                self.logger.warning(
                    "PID {} belongs to zombie process. "
                    "This means that parent process is still alive, while "
                    "child process have been killed. This is OK during tests, "
                    "but should never happen in real life.".format(self.pid)
                )
                is_alive = False
            else:
                is_alive = True
        else:
            is_alive = process_exists  # False or None (no access)
        return is_alive

    def terminate(self):
        """Terminates the process"""
        return self._kill(signal.SIGTERM)

    def kill(self):
        """Kills the process"""
        return self._kill(signal.SIGKILL)

    @pid_required(otherwise="")
    def status(self):
        """Retrieves status of the process using `ps` command"""
        try:
            stat = force_text(
                subprocess.check_output(
                    ["ps", "-p", str(self.pid), "-o", "stat="]
                ).strip()
            )
        except subprocess.CalledProcessError:
            stat = ""
        return stat



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

import time

from django.core.management import BaseCommand, CommandError

from cronman.remote_manager import CronRemoteManager
from cronman.scheduler import CronScheduler


class Command(BaseCommand):
    help = (
        "Cron Manager: CLI to control remote Cron Schedulers via Redis.\n"
        "Redis access configuration have to be the same on local and remote "
        "machines.\n"
        "With default configuration, the following settings have to match:\n"
        "CRONMAN_REDIS_HOST, CRONMAN_REDIS_PORT, CRONMAN_REDIS_DB.\n"
    )

    ALLOWED_METHODS = (
        "disable",
        "enable",
        "get_status",
        "clear_status",
        "kill",
    )

    def add_arguments(self, parser):
        parser.add_argument(
            "method",
            help=(
                "Choices: disable, enable, get_status, clear_status, "
                "kill:<jobspec>."
            ),
        )
        parser.add_argument("hosts", nargs="+")
        parser.add_argument(
            "--wait",
            action="store_true",
            default=False,
            help="Wait for Cron Schedulers to receive the request",
        )

    def handle(self, **options):
        """Main command logic"""
        method_string = options["method"]
        if ":" in method_string:
            method_name, arg = method_string.split(":", 1)
            args = (arg,)
        else:
            method_name = method_string
            args = ()
        if method_name not in self.ALLOWED_METHODS:
            raise CommandError(
                'Command "{}" is not allowed'.format(method_name)
            )
        host_names = options["hosts"]
        remote_manager = CronRemoteManager()
        method = getattr(remote_manager, method_name)
        results = []
        for host_name in host_names:
            result = method(*args, host_name=host_name)
            results.append(
                "{} {} -> {}".format(method_string, host_name, result)
            )
        summary = "\n".join(results)
        if options["wait"]:
            time.sleep(CronScheduler.interval * 60)
        return summary



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from cronman.worker.cron_job_info import CronJobClassList
from cronman.worker.process_manager import ProcessManager
from cronman.worker.signal_notifier import SignalNotifier
from cronman.worker.worker import CronWorker
from cronman.worker.worker_file import CronWorkerPIDFile
from cronman.worker.worker_list import CronWorkerPIDList

__all__ = [
    "CronJobClassList",
    "CronWorker",
    "CronWorkerPIDFile",
    "CronWorkerPIDList",
    "ProcessManager",
    "SignalNotifier",
]



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

import os
import platform
import tempfile

from django.conf import settings

MYPY = False
if MYPY:
    from typing import Dict, Optional, Text


UNDEFINED = object()


class Setting(object):
    """AppSettings property."""

    def __init__(self, name, default):
        self.name = name
        self.default = default

    def __get__(self, instance, cls=None):
        if instance is None:
            return self
        return getattr(settings, self.name, self.default)


class AppSettings(object):
    """Class providing access to optional `cronman` app settings."""

    # General app settings:
    CRONMAN_DATA_DIR = Setting(
        "CRONMAN_DATA_DIR", os.path.join(tempfile.gettempdir(), "test.cronman")
    )  # type: Text
    CRONMAN_DEBUG = Setting("CRONMAN_DEBUG", False)  # type: bool
    # Module responsible for cron scheduler configuration,
    # the list of tasks available to select (admin)
    # and execute (cron machines):
    CRONMAN_JOBS_MODULE = Setting(
        "CRONMAN_JOBS_MODULE", None
    )  # type: Optional[Text]
    CRONMAN_NICE_CMD = Setting("CRONMAN_NICE_CMD", "nice")  # type: Text
    CRONMAN_IONICE_CMD = Setting(
        "CRONMAN_IONICE_CMD",
        "ionice" if platform.system() == "Linux" else None,
    )  # type: Optional[Text]
    CRONMAN_REMOTE_MANAGER_ENABLED = Setting(
        "CRONMAN_REMOTE_MANAGER_ENABLED", False
    )  # type: bool
    CRONMAN_RUN_CRON_TASKS_CRONITOR_ID = Setting(
        "CRONMAN_RUN_CRON_TASKS_CRONITOR_ID", None
    )  # type: Optional[Text]
    CRONMAN_CRON_SCHEDULER_CRONITOR_ID = Setting(
        "CRONMAN_CRON_SCHEDULER_CRONITOR_ID", None
    )  # type: Optional[Text]
    CRONMAN_CLEAN_CRON_TASKS_CRONITOR_ID = Setting(
        "CRONMAN_CLEAN_CRON_TASKS_CRONITOR_ID", None
    )  # type: Optional[Text]
    CRONMAN_ADMIN_SITE = Setting(
        "CRONMAN_ADMIN_SITE", "django.contrib.admin.site"
    )  # type: Optional[Text]

    # Slack notifications settings:
    CRONMAN_SLACK_ENABLED = Setting(
        "CRONMAN_SLACK_ENABLED", False
    )  # type: bool
    CRONMAN_SLACK_URL = Setting(
        "CRONMAN_SLACK_URL", None
    )  # type: Optional[Text]
    CRONMAN_SLACK_TOKEN = Setting(
        "CRONMAN_SLACK_TOKEN", None
    )  # type: Optional[Text]
    CRONMAN_SLACK_DEFAULT_CHANNEL = Setting(
        "CRONMAN_SLACK_DEFAULT_CHANNEL", None
    )  # type: Optional[Text]

    # Cronitor settings:
    CRONMAN_CRONITOR_ENABLED = Setting(
        "CRONMAN_CRONITOR_ENABLED", False
    )  # type: bool
    CRONMAN_CRONITOR_URL = Setting(
        "CRONMAN_CRONITOR_URL",
        "https://cronitor.link/{cronitor_id}/{end_point}",
    )  # type: Text

    # Sentry configuration:
    CRONMAN_SENTRY_ENABLED = Setting(
        "CRONMAN_SENTRY_ENABLED", False
    )  # type: bool
    CRONMAN_SENTRY_CONFIG = Setting(
        "CRONMAN_SENTRY_CONFIG", {"dsn": ""}
    )  # type: Dict
    # Absolute path to raven-cmd script (Sentry wrapper):
    CRONMAN_RAVEN_CMD = Setting(
        "CRONMAN_RAVEN_CMD", None
    )  # type: Optional[Text]

    # Redis configuration:
    CRONMAN_REDIS_HOST = Setting(
        "CRONMAN_REDIS_HOST", "127.0.0.1"
    )  # type: Text
    CRONMAN_REDIS_PORT = Setting("CRONMAN_REDIS_PORT", 6379)  # type: int
    CRONMAN_REDIS_DB = Setting("CRONMAN_REDIS_DB", 0)  # type: int
    CRONMAN_REDIS_CONSTRUCTOR = Setting(
        "CRONMAN_REDIS_CONSTRUCTOR",
        "cronman.redis_client.get_strict_redis_default",
    )  # type: Text


app_settings = AppSettings()



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

from django.test import override_settings
from django.utils import timezone

from cronman.forms import CronTaskAdminForm
from cronman.models import CronTask
from cronman.tests import cron_jobs
from cronman.tests.base import BaseCronTestCase


class CronTaskAdminFormTestCase(BaseCronTestCase):
    """Tests for CronTaskAdminForm"""

    @override_settings(CRONMAN_JOBS_MODULE=None)
    def test_cron_job_choices_empty(self):
        """Test that cron job choices are empty when CRONMAN_JOBS_MODULE is not
        set.
        """
        form = CronTaskAdminForm(instance=CronTask())
        self.assertEqual(form.fields["cron_job"].choices, [])

    @override_settings(CRONMAN_JOBS_MODULE="cronman.tests.cron_jobs")
    def test_cron_job_choices_matching(self):
        """Test that cron job choices are matching CRONMAN_JOBS_MODULE
        configuration.
        """
        form = CronTaskAdminForm(instance=CronTask())
        self.assertEqual(
            set(
                value
                for value, verbose_name in form.fields["cron_job"].choices
            ),
            set(cron_jobs.ALLOWED_CRON_TASKS),
        )

    @override_settings(CRONMAN_JOBS_MODULE="cronman.tests.cron_jobs")
    def test_valid_form_no_params(self):
        """Test for valid form - no cron job params"""
        now = timezone.now()
        cron_job = "Sleep"
        form = CronTaskAdminForm(
            data={"cron_job": cron_job, "start_at": now}, instance=CronTask()
        )
        self.assertTrue(form.is_valid())
        self.assertEqual(
            form.cleaned_data,
            {"cron_job": cron_job, "params": "", "start_at": now},
        )

    @override_settings(CRONMAN_JOBS_MODULE="cronman.tests.cron_jobs")
    def test_valid_form_with_params(self):
        """Test for valid form - valid cron job params"""
        now = timezone.now()
        cron_job = "Sleep"
        params = 'makers="Toyota;Audi"'
        form = CronTaskAdminForm(
            data={"cron_job": cron_job, "params": params, "start_at": now},
            instance=CronTask(),
        )
        self.assertTrue(form.is_valid())
        self.assertEqual(
            form.cleaned_data,
            {"cron_job": cron_job, "params": params, "start_at": now},
        )

    @override_settings(CRONMAN_JOBS_MODULE="cronman.tests.cron_jobs")
    def test_invalid_params(self):
        """Test for invalid form - invalid cron job params"""
        now = timezone.now()
        cron_job = "Sleep"
        params = 'makers="Toyota;Audi",1234'  # params syntax error
        form = CronTaskAdminForm(
            data={"cron_job": cron_job, "params": params, "start_at": now},
            instance=CronTask(),
        )
        self.assertFalse(form.is_valid())



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

import logging
import platform

from cronman.monitor import Cronitor, Sentry, Slack
from cronman.utils import bool_param, config, ensure_dir, format_exception

logger = logging.getLogger("cronman.command")


class BaseCronObject(object):
    """Common base class for CronRemoteManager, CronScheduler, CronSpawner,
    CronWorker.
    """

    def __init__(self, **kwargs):
        self.data_dir = kwargs.get("data_dir", config("CRONMAN_DATA_DIR"))
        self.debug = kwargs.get("debug", bool_param(config("CRONMAN_DEBUG")))
        self.cronitor = Cronitor()
        self.sentry = Sentry()
        self.slack = Slack()
        ensure_dir(self.data_dir)
        self.logger = kwargs.get("logger", logger)

    def warning(self, exception, silent=False):
        """Handles exception as warning"""
        message = format_exception(exception)
        if not silent:
            self.logger.warning(message)
            system_name = platform.node()
            self.slack.post(
                "[{host}] {message}".format(host=system_name, message=message)
            )
        return message + "\n"  # to be printed on stdout



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

import os

import mock

from cronman import utils
from cronman.tests.base import TEST_CRONMAN_DATA_DIR, BaseCronTestCase


class CronUtilsTestCase(BaseCronTestCase):
    """Tests for `cronman.utils` module"""

    def test_function_signature(self):
        """Test for `function_signature` function"""

        def func_a(a, b, c=42, d=None, *args, **kwargs):
            pass

        def func_b():
            pass

        self.assertEqual(
            utils.function_signature(func_a),
            "a, b, c=42, d=None, *args, **kwargs",
        )
        self.assertEqual(utils.function_signature(func_b), "")

    def test_function_signature_method(self):
        """Test for `function_signature` function - case: method"""

        class X(object):
            def method(self, a, b, c=42, d=None, *args, **kwargs):
                pass

        self.assertEqual(
            utils.function_signature(X.method),
            "a, b, c=42, d=None, *args, **kwargs",
        )

    def test_parse_params_empty(self):
        """Test for `parse_params` function - case: empty string"""
        self.assertEqual(utils.parse_params(""), ([], {}))

    def test_parse_params_positional(self):
        """Test for `parse_params` function - case: positional arguments"""
        self.assertEqual(
            utils.parse_params("Test, 1,4.2 ,\"A; B; C\", '386'"),
            (["Test", "1", "4.2", "A; B; C", "386"], {}),
        )

    def test_parse_params_named(self):
        """Test for `parse_params` function - case: keyword arguments"""
        self.assertEqual(
            utils.parse_params("a=Test, b=4.2 ,p_1=\"A; B; C\", TEST='386'"),
            ([], {"a": "Test", "b": "4.2", "p_1": "A; B; C", "TEST": "386"}),
        )

    def test_parse_params_mixed(self):
        """Test for `parse_params` function - case: mixed arguments"""
        self.assertEqual(
            utils.parse_params("Test, 1,4.2 ,p_1=\"A; B; C\", TEST='386'"),
            (["Test", "1", "4.2"], {"p_1": "A; B; C", "TEST": "386"}),
        )

    def test_parse_params_mixed_quoted(self):
        """Test for `parse_params` function - case: mixed arguments with
        quoted values (comma, space)
        """
        self.assertEqual(
            utils.parse_params("\"a b\",'1,2',a='a b',b=\"1,2\""),
            (["a b", "1,2"], {"a": "a b", "b": "1,2"}),
        )

    def test_parse_params_invalid_implicit_empty_value(self):
        """Test for `parse_params` function - case: invalid params -
        implicit empty value
        """
        with self.assertRaisesMessage(
            ValueError,
            "In chars 5-11 ` TEST=`: "
            'Implicit empty value. Use explicit `""` instead.',
        ):
            utils.parse_params("Test, TEST=")

    def test_parse_params_invalid_unbalanced_parentheses(self):
        """Test for `parse_params` function - case: invalid params -
        unbalanced parentheses
        """
        with self.assertRaisesMessage(
            ValueError, 'In chars 5-12 ` TEST="`: Unbalanced parentheses.'
        ):
            utils.parse_params('Test, TEST="')

    def test_parse_params_invalid_positional_after_named_arg(self):
        """Test for `parse_params` function - case: invalid params -
        positional argument after named one
        """
        with self.assertRaisesMessage(
            ValueError,
            "In chars 13-18 `4.2 ,`: "
            "Positional argument after named arguments.",
        ):
            utils.parse_params("Test, TEST=1,4.2 ,\"A; B; C\", '386'")

    def test_parse_params_invalid_duplicated_named_arg(self):
        """Test for `parse_params` function - case: invalid params -
        duplicated named argument
        """
        with self.assertRaisesMessage(
            ValueError, "In chars 8-15 `param=2`: Duplicated named argument."
        ):
            utils.parse_params("param=1,param=2")

    def test_parse_job_spec_no_params(self):
        """Test for `parse_job_spec` function - case: no params"""
        self.assertEqual(utils.parse_job_spec("CronJob"), ("CronJob", [], {}))

    def test_parse_job_spec_valid_params(self):
        """Test for `parse_job_spec` function - case: valid params"""
        self.assertEqual(
            utils.parse_job_spec('CronJob:42,a=1, b="a: b c"'),
            ("CronJob", ["42"], {"a": "1", "b": "a: b c"}),
        )

    def test_parse_job_spec_invalid_params(self):
        """Test for `parse_job_spec` function - case: invalid params"""
        with self.assertRaisesMessage(
            ValueError, "Positional argument after named arguments."
        ):
            utils.parse_job_spec('CronJob:a=1,"test"'),

    def test_bool_param_true(self):
        """Test for `bool_param` function - case: true values"""
        self.assertTrue(utils.bool_param("1"))
        self.assertTrue(utils.bool_param("Yes"))
        self.assertTrue(utils.bool_param("y"))
        self.assertTrue(utils.bool_param("True"))
        self.assertTrue(utils.bool_param("ON"))

    def test_bool_param_false(self):
        """Test for `bool_param` function - case: false values"""
        self.assertFalse(utils.bool_param("0"))
        self.assertFalse(utils.bool_param("no"))
        self.assertFalse(utils.bool_param("N"))
        self.assertFalse(utils.bool_param("false"))
        self.assertFalse(utils.bool_param("Off"))

    def test_bool_param_undefined(self):
        """Test for `bool_param` function - case: undefined values"""
        self.assertIsNone(utils.bool_param(""))
        self.assertIsNone(utils.bool_param(None))

    def test_bool_param_no_conversion(self):
        """Test for `bool_param` function - case: no conversion"""
        self.assertTrue(utils.bool_param(True))
        self.assertFalse(utils.bool_param(False))

    def test_flag_param_explicit_value_true(self):
        """Test for `flag_param` function - case: explicit value, true"""
        self.assertTrue(utils.flag_param("yes", "f1", None))
        self.assertTrue(utils.flag_param(True, "f1", ""))

    def test_flag_param_explicit_value_false(self):
        """Test for `flag_param` function - case: explicit value, false"""
        self.assertFalse(utils.flag_param("no", "f1", ""))
        self.assertFalse(utils.flag_param(False, "f1", None))

    def test_flag_param_fallback_value_true(self):
        """Test for `flag_param` function - case: explicit value not provided -
        fallback to flags, true (flag found)
        """
        self.assertTrue(utils.flag_param(None, "f1", "f1 f2"))
        self.assertTrue(utils.flag_param(None, "f1", "-f5-f1"))
        self.assertTrue(utils.flag_param("", "f1", "f0,f1,f2"))
        self.assertTrue(utils.flag_param("", "f1", "f0;f1;f2"))

    def test_flag_param_fallback_value_false(self):
        """Test for `flag_param` function - case: explicit value not provided -
        fallback to flags, false (flag not found)
        """
        self.assertFalse(utils.flag_param(None, "f1", "f3 f2"))
        self.assertFalse(utils.flag_param(None, "f1", "-f5-f3"))
        self.assertFalse(utils.flag_param(None, "f1", ""))
        self.assertFalse(utils.flag_param(None, "f1", None))
        self.assertFalse(utils.flag_param("", "f1", "f0,f3,f2"))
        self.assertFalse(utils.flag_param("", "f1", "f0;f3;f2"))
        self.assertFalse(utils.flag_param("", "f1", ""))
        self.assertFalse(utils.flag_param("", "f1", None))

    def test_list_param_default(self):
        """Test for `list_param` function - case: default values"""
        self.assertEqual(utils.list_param("a B 42"), ["a", "B", "42"])
        self.assertEqual(utils.list_param("a\tXY\t1.5\t"), ["a", "XY", "1.5"])
        self.assertEqual(utils.list_param(""), [])

    def test_list_param_custom_integers(self):
        """Test for `list_param` function - case: customized params,
        semicolon-separated list of integers
        """
        self.assertEqual(
            utils.list_param("12; 2; 386", delimiter=";", cast=int),
            [12, 2, 386],
        )

    def test_list_param_custom_string(self):
        """Test for `list_param` function - case: customized params,
        tab-separated list of strings, keep empty items, don't strip spaces
        """
        self.assertEqual(
            utils.list_param(
                "Text 1 \tTEST\t B \t",
                delimiter="\t",
                skip_empty=False,
                strip=False,
            ),
            ["Text 1 ", "TEST", " B ", ""],
        )

    def test_list_param_custom_default_and_replace(self):
        """Test for `list_param` function - case: default and replace_map"""
        self.assertEqual(
            utils.list_param("", default=["default"]), ["default"]
        )
        self.assertEqual(
            utils.list_param("A B X", replace_map={"X": "Y"}), ["A", "B", "Y"]
        )

    def test_ensure_dir_existing_directory(self):
        """Test for `ensure_dir` function - case: existing directory"""
        os.makedirs(TEST_CRONMAN_DATA_DIR)
        self.assertTrue(utils.ensure_dir(TEST_CRONMAN_DATA_DIR))

    def test_ensure_dir_new_directory_tree(self):
        """Test for `ensure_dir` function - case: new directory tree"""
        self.assertTrue(
            utils.ensure_dir(os.path.join(TEST_CRONMAN_DATA_DIR, "a/b/c"))
        )

    def test_ensure_dir_new_directory_failure(self):
        """Test for `ensure_dir` function - case: it's a file"""
        os.makedirs(TEST_CRONMAN_DATA_DIR)
        path = os.path.join(TEST_CRONMAN_DATA_DIR, "text.txt")
        with open(path, "a"):
            self.assertFalse(utils.ensure_dir(path))

    def test_spawn(self):
        """Test for `spawn` function"""
        with mock.patch("cronman.utils.subprocess.Popen") as mock_popen:
            type(mock_popen.return_value).pid = mock.PropertyMock(
                return_value=42
            )
            self.assertEqual(utils.spawn("ls", "-la"), 42)

    def test_execute(self):
        """Test for `execute` function"""
        path = os.path.join(TEST_CRONMAN_DATA_DIR, "execute")
        os.makedirs(path)
        with open(os.path.join(path, "aaa.txt"), "a"):
            pass
        with open(os.path.join(path, "bbb.txt"), "a"):
            pass
        result = utils.execute("ls", path)
        self.assertTrue(result.ok)
        self.assertEqual(result.status, 0)
        self.assertEqual(result.args, ("ls", path))
        self.assertEqual(result.command, "ls {}".format(path))
        self.assertEqual(
            set(result.output_text.strip().split("\n")), {"aaa.txt", "bbb.txt"}
        )

    def test_execute_pipe(self):
        """Test for `execute_pipe` function"""
        path = os.path.join(TEST_CRONMAN_DATA_DIR, "execute")
        os.makedirs(path)
        with open(os.path.join(path, "aaa.txt"), "a"):
            pass
        with open(os.path.join(path, "bbb.txt"), "a"):
            pass
        with open(os.path.join(path, "abb.txt"), "a"):
            pass
        result = utils.execute_pipe(["ls", "-A1", path], ["grep", "bb"])
        self.assertTrue(result.ok)
        self.assertEqual(result.status, 0)
        self.assertEqual(result.args, (["ls", "-A1", path], ["grep", "bb"]))
        self.assertEqual(result.command, "ls -A1 {} | grep bb".format(path))
        self.assertEqual(
            set(result.output_text.strip().split("\n")), {"abb.txt", "bbb.txt"}
        )

    def test_execute_pipe_non_zero_return_code(self):
        """Test for `execute_pipe` function - case: non-zero return code"""
        path = os.path.join(TEST_CRONMAN_DATA_DIR, "execute")
        os.makedirs(path)
        result = utils.execute_pipe(["ls", "-A1", path], ["grep", "bb"])
        self.assertFalse(result.ok)
        self.assertEqual(result.status, 1)

    def test_cwd_ok(self):
        """Test for `CWD` context manager - case: no exception raised"""
        path = os.path.realpath(os.path.join(TEST_CRONMAN_DATA_DIR, "cwd"))
        os.makedirs(path)
        cwd = os.getcwd()
        with utils.CWD(path):
            self.assertEqual(os.getcwd(), path)
            self.assertEqual(utils.execute("pwd").output_text.strip(), path)
        self.assertEqual(os.getcwd(), cwd)

    def test_cwd_exception(self):
        """Test for `CWD` context manager - case: exception raised"""
        path = os.path.realpath(os.path.join(TEST_CRONMAN_DATA_DIR, "cwd"))
        os.makedirs(path)
        cwd = os.getcwd()
        try:
            with utils.CWD(path):
                self.assertEqual(os.getcwd(), path)
                self.assertEqual(
                    utils.execute("pwd").output_text.strip(), path
                )
                raise RuntimeError
        except RuntimeError:
            pass
        self.assertEqual(os.getcwd(), cwd)

    @mock.patch("cronman.utils.os.environ", {"CRON_PROCESS_RESUMED": "1"})
    def test_is_cron_process_resumed_yes(self):
        """Test for `is_cron_process_resumed` function - case: yes"""
        self.assertTrue(utils.is_cron_process_resumed())

    @mock.patch("cronman.utils.os.environ", {"CRON_PROCESS_RESUMED": "0"})
    def test_is_cron_process_resumed_no(self):
        """Test for `is_cron_process_resumed` function - case: no"""
        self.assertFalse(utils.is_cron_process_resumed())

    @mock.patch("cronman.utils.os.environ", {})
    def test_is_cron_process_resumed_unset(self):
        """Test for `is_cron_process_resumed` function - case: unset (no)"""
        self.assertFalse(utils.is_cron_process_resumed())

    @mock.patch("cronman.utils.os.environ", {"CRON_PROCESS_RESUMED": ""})
    def test_is_cron_process_resumed_empty(self):
        """Test for `is_cron_process_resumed` function - case: empty (no)"""
        self.assertFalse(utils.is_cron_process_resumed())

    @mock.patch("cronman.worker.worker.CronWorker.get_pid_file")
    @mock.patch("cronman.worker.worker.CronWorker.pid_file_locked")
    def test_is_cron_job_running_default(
        self, mock_pid_file_locked, mock_get_pid_file
    ):
        """Test for `is_cron_job_running` function - case: default params"""
        mock_pid_file = mock.MagicMock()
        mock_get_pid_file.return_value = mock_pid_file
        mock_pid_file_locked.return_value = True

        class FakeCronJob(object):
            pass

        self.assertTrue(utils.is_cron_job_running(FakeCronJob))

        mock_get_pid_file.assert_called_once_with(
            cron_job_class=FakeCronJob, name="FakeCronJob", args=[], kwargs={}
        )
        mock_pid_file_locked.assert_called_once_with(mock_pid_file, 1)

    @mock.patch("cronman.worker.worker.CronWorker.get_pid_file")
    @mock.patch("cronman.worker.worker.CronWorker.pid_file_locked")
    def test_is_cron_job_running_custom(
        self, mock_pid_file_locked, mock_get_pid_file
    ):
        """Test for `is_cron_job_running` function - case: custom params"""
        mock_pid_file = mock.MagicMock()
        mock_get_pid_file.return_value = mock_pid_file
        mock_pid_file_locked.return_value = False

        class FakeCronJob(object):
            pass

        self.assertFalse(
            utils.is_cron_job_running(
                FakeCronJob,
                name="AlternativeName",
                args=[42],
                kwargs={"param": 0},
            )
        )

        mock_get_pid_file.assert_called_once_with(
            cron_job_class=FakeCronJob,
            name="AlternativeName",
            args=[42],
            kwargs={"param": 0},
        )
        mock_pid_file_locked.assert_called_once_with(mock_pid_file, 1)



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

import logging
import socket

from django.utils.functional import cached_property

from six.moves import range  # pylint: disable=E0401, E0611

from cronman.base import BaseCronObject
from cronman.exceptions import MissingDependency
from cronman.redis_client import get_strict_redis
from cronman.taxonomies import CronSchedulerStatus
from cronman.utils import bool_param, config

logger = logging.getLogger("cronman.command.cron_remote_manager")


class CronRemoteManager(BaseCronObject):
    """Cron Manager: controls remote Cron Schedulers via Redis."""

    STATUS_KEY = "cron_scheduler:status:{host_name}"
    KILL_KEY = "cron_scheduler:kill:{host_name}"

    MAX_KILLS = 5

    def __init__(self, host_name=None, **kwargs):
        kwargs["logger"] = kwargs.get("logger", logger)
        super(CronRemoteManager, self).__init__(**kwargs)
        self.host_name = host_name or socket.gethostname()
        self.enabled = bool_param(config("CRONMAN_REMOTE_MANAGER_ENABLED"))

    # Redis operations:

    @cached_property
    def redis_client(self):
        """Redis client object (StrictRedis)"""
        return get_strict_redis()

    def _redis_call(self, method_name, args, description):
        """Wrapper over Redis client method call"""
        if self.enabled:
            try:
                from redis import ConnectionError
            except ImportError:
                raise MissingDependency(
                    "Unable to import redis. "
                    "CronRemoteManager requires this dependency."
                )
            try:
                method = getattr(self.redis_client, method_name)
                result = method(*args)
            except ConnectionError as error:
                self.logger.warning(
                    "Remote Manager: {} FAILED: {}".format(description, error)
                )
                result = None
            else:
                self.logger.info(
                    "Remote Manager: {} OK: {}".format(description, result)
                )
        else:
            self.logger.warning(
                "Remote Manager: {} CANCELLED: "
                "disabled in configuration".format(description)
            )
            result = None
        return result

    def redis_set(self, key, value):
        """Sets key's value in Redis"""
        return self._redis_call(
            "set", (key, value), "SET {}={}".format(key, value)
        )

    def redis_get(self, key):
        """Retrieves key's value from Redis"""
        return self._redis_call("get", (key,), "GET {}".format(key))

    def redis_delete(self, key):
        """Removes key from Redis"""
        return self._redis_call("delete", (key,), "DEL {}".format(key))

    def redis_rpush(self, key, value):
        """Appends value to the list on given key in Redis"""
        return self._redis_call(
            "rpush", (key, value), "RPUSH {} {}".format(key, value)
        )

    def redis_lpop(self, key):
        """Pops first element from list on given key in Redis"""
        return self._redis_call("lpop", (key,), "LPOP {}".format(key))

    # Redis keys:

    def get_status_key(self, host_name=None):
        """Creates Cron Scheduler status key from host name"""
        return self.STATUS_KEY.format(host_name=host_name or self.host_name)

    def get_kill_key(self, host_name=None):
        """Creates Cron Scheduler kill command key from host name"""
        return self.KILL_KEY.format(host_name=host_name or self.host_name)

    # Cron Scheduler status operations:

    def set_status(self, status, host_name=None):
        """Sets Cron Scheduler status"""
        return self.redis_set(self.get_status_key(host_name=host_name), status)

    def get_status(self, host_name=None):
        """Retrieves Cron Scheduler status"""
        return self.redis_get(self.get_status_key(host_name=host_name))

    def clear_status(self, host_name=None):
        """Removes Cron Scheduler status"""
        return self.redis_delete(self.get_status_key(host_name=host_name))

    def pop_status(self, host_name=None):
        """Retrieves and removes Cron Scheduler status"""
        status = self.get_status(host_name=host_name)
        if status is not None:
            self.clear_status(host_name=host_name)
        return status

    # Cron Scheduler kill command operations:

    def pop_killed(self, host_name=None):
        """Pops all cron jobs to be killed"""
        key = self.get_kill_key(host_name=host_name)
        job_specs = set()
        for i in range(self.MAX_KILLS):
            job_spec = self.redis_lpop(key)
            if job_spec:
                job_specs.add(job_spec)
            else:
                break
        return job_specs

    def kill(self, job_spec, host_name=None):
        """Ask the scheduler to kill a cron job of given job_spec"""
        return self.redis_rpush(
            self.get_kill_key(host_name=host_name), job_spec
        )

    # Shortcuts:

    def disable(self, host_name=None):
        """Ask the scheduler to disable itself and kill running workers"""
        return self.set_status(
            status=CronSchedulerStatus.DISABLED, host_name=host_name
        )

    def enable(self, host_name=None):
        """Ask the scheduler to enable itself and resume killed workers"""
        return self.set_status(
            status=CronSchedulerStatus.ENABLED, host_name=host_name
        )



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

import importlib
import os
import re
from distutils import log
from distutils.errors import DistutilsError

from setuptools import Command, find_packages, setup
from setuptools.command.test import test as BaseTestCommand

DJANGO_SETTINGS_MODULE = "cronman.tests.settings"


def read(*path_parts):
    """Retrieve content of a text file"""
    file_path = os.path.join(os.path.dirname(__file__), *path_parts)
    with open(file_path) as file_obj:
        return file_obj.read()


def find_version(*path_parts):
    """Retrieve current version string"""
    version_file_contents = read(*path_parts)
    version_match = re.search(
        r'^__version__ = ["\'](?P<version>[^"\']*)["\']',
        version_file_contents,
        re.M,
    )
    if not version_match:
        raise RuntimeError("Unable to find version string.")
    return version_match.group("version")


def django_setup(debug=False):
    """Set up Django using test settings module."""
    os.environ["DJANGO_SETTINGS_MODULE"] = DJANGO_SETTINGS_MODULE
    if debug:
        os.environ.setdefault("TEST_CRONMAN_LOG_LEVEL", "DEBUG")

    import django

    django.setup()


class TestCommand(BaseTestCommand):
    """Command to run unit tests (both pure-Python and Django) after in-place
    build.

    # NOTE: Pure-Python unit tests disabled.
    """

    def run_tests(self):
        # Pure-Python unit tests from `tests/*` - disabled.
        # BaseTestCommand.run_tests(self)

        # Run Django tests:

        django_setup()

        from django.test.utils import get_runner

        test_runner_class = get_runner(
            importlib.import_module(DJANGO_SETTINGS_MODULE),
            test_runner_class="django.test.runner.DiscoverRunner",
        )

        test_runner = test_runner_class(verbosity=2, interactive=True)
        failures = test_runner.run_tests(["cronman"])
        if failures:
            msg = "Django tests failed."
            self.announce(msg, log.ERROR)
            raise DistutilsError(msg)


class BaseAliasCommand(Command):
    """Setup command working as simple alias to Django's command."""

    user_options = []

    django_command = "help"

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        """Run `shell`"""
        django_setup(debug=True)

        from django.core.management import call_command

        call_command(self.django_command)


class ShellCommand(BaseAliasCommand):
    """Command to run Django's `shell`."""

    description = "run Django's `shell`."

    django_command = "shell"


class MakeMigrationsCommand(BaseAliasCommand):
    """Command to run Django's `makemigrations`."""

    description = "run Django's `makemigrations`."

    django_command = "makemigrations"


class MigrateCommand(BaseAliasCommand):
    """Command to run Django's `migrate`."""

    description = "run Django's `migrate`."

    django_command = "migrate"


setup(
    name="django-cronman",
    version=find_version("cronman", "version.py"),
    description="Cron management app for Django",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    url="https://github.com/unhaggle/django-cronman",
    author="Unhaggle Inc.",
    author_email="",
    packages=find_packages(),
    install_requires=[
        "croniter < 0.4",
        "django",
        "python-dateutil",
        "requests",
        "six",
        "typing ; python_version < '3.0'",
    ],
    include_package_data=True,
    tests_require=["mock", "sentry_sdk", "redis < 2.11"],
    extras_require={"redis": ["redis < 2.11"], "sentry": ["sentry_sdk"]},
    cmdclass={
        "test": TestCommand,
        "shell": ShellCommand,
        "makemigrations": MakeMigrationsCommand,
        "migrate": MigrateCommand,
    },
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: BSD License",
        "Framework :: Django",
        "Framework :: Django :: 1.11",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
        "Environment :: Web Environment",
        "Operating System :: POSIX",
        "Topic :: Utilities",
    ],
)



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

import logging

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.functional import cached_property
from django.utils.html import strip_tags
from django.utils.http import urlencode

import requests
from six.moves.html_parser import HTMLParser

from cronman.config import app_settings
from cronman.exceptions import MissingDependency
from cronman.utils import bool_param, chunks, config, format_exception

logger = logging.getLogger("cronman.command")


def _get_sentry_sdk():
    """Creates raven.Client instance configured to work with cron jobs."""
    # NOTE: this function uses settings and therefore it shouldn't be called
    #       at module level.
    try:
        sentry_sdk = __import__("sentry_sdk")
        DjangoIntegration = __import__(
            "sentry_sdk.integrations.django"
        ).integrations.django.DjangoIntegration
    except ImportError:
        raise MissingDependency(
            "Unable to import sentry_sdk. "
            "Sentry monitor requires this dependency."
        )

    for setting in (
        "CRONMAN_SENTRY_CONFIG",
        "SENTRY_CONFIG",
        "RAVEN_CONFIG",
    ):
        client_config = getattr(settings, setting, None)
        if client_config is not None:
            break
    else:
        client_config = app_settings.CRONMAN_SENTRY_CONFIG

    sentry_sdk.init(integrations=[DjangoIntegration()], **client_config)
    return sentry_sdk


class Cronitor(object):
    """Wrapper over Cronitor web API"""

    def __init__(self):
        self.logger = logger
        self.enabled = bool_param(config("CRONMAN_CRONITOR_ENABLED"))
        self.url = app_settings.CRONMAN_CRONITOR_URL

    def run(self, cronitor_id, msg=None):
        """Pings Cronitor when job started"""
        self._ping(cronitor_id, "run", msg)

    def complete(self, cronitor_id, msg=None):
        """Pings Conitor when job is done"""
        self._ping(cronitor_id, "complete", msg)

    def fail(self, cronitor_id, msg=None):
        """Pings Cronitor when job failed"""
        self._ping(cronitor_id, "fail", msg)

    def _ping(self, cronitor_id, end_point, msg):
        """Sends a request to Cronitor web API"""
        if not self.enabled:
            self.logger.warning(
                "Cronitor request ignored (disabled in settings)."
            )
            return
        url = self.url.format(cronitor_id=cronitor_id, end_point=end_point)
        params = {"msg": msg} if msg else None
        try:
            response = requests.head(url, params=params, timeout=10)
            response.raise_for_status()
        except requests.RequestException as error:
            # Catch all network and HTTP errors raised by `requests`
            self.logger.warning(
                "Cronitor request failed: {} {}".format(
                    url, format_exception(error)
                )
            )


class Sentry(object):
    """Wrapper over Sentry API"""

    def __init__(self):
        self.logger = logger
        self.enabled = bool_param(config("CRONMAN_SENTRY_ENABLED"))
        self.raven_cmd = app_settings.CRONMAN_RAVEN_CMD

    @cached_property
    def _sentry_sdk(self):
        return _get_sentry_sdk()

    def capture_exception(self, exception):
        if self.enabled:
            self._sentry_sdk.capture_exception(exception)
        else:
            self.logger.debug("Sentry request ignored (disabled in settings).")


def send_errors_to_sentry(method):
    """Decorator for CronScheduler / CronWorker public methods aimed to send
    unhandled exceptions to Sentry.
    """

    def _method(self, *args, **kwargs):
        try:
            return method(self, *args, **kwargs)
        except Exception as e:
            self.sentry.capture_exception(e)
            raise

    return _method


class Slack(object):
    """Wrapper over Slack web API"""

    def __init__(self):
        self.logger = logger
        self.enabled = bool_param(config("CRONMAN_SLACK_ENABLED"))
        self.url = app_settings.CRONMAN_SLACK_URL
        self.token = app_settings.CRONMAN_SLACK_TOKEN
        self.default_channel = app_settings.CRONMAN_SLACK_DEFAULT_CHANNEL

    def _prepare_message(self, message):
        # slack don't process html entities
        html_parser = HTMLParser()
        message = html_parser.unescape(message)
        # slack also don't render html itself
        message = strip_tags(message)
        return message

    def post(self, message, channel=None):
        """Posts a message to Slack channel"""
        if not self.enabled:
            self.logger.warning(
                "Slack request ignored (disabled in settings)."
            )
            return
        if not (self.url and self.token):
            raise ImproperlyConfigured(
                "CRONMAN_SLACK_URL and CRONMAN_SLACK_TOKEN are required by "
                "Slack integration. Please provide values for these settings "
                "or disable Slack integration (CRONMAN_SLACK_ENABLED = False)."
            )
        channel_url = "{}?{}".format(
            self.url,
            urlencode(
                (
                    ("token", self.token),
                    ("channel", "#{}".format(channel or self.default_channel)),
                )
            ),
        )
        # prepare message
        message = self._prepare_message(message)

        try:
            # We need to split message into packages because of
            # openssl/urllib3 error. We are going for 12000 chars here
            # as tests showed that this is properly sent to Slack
            chunked_message = chunks(message, 12000)
            for chunk in chunked_message:
                # Sending data of type unicode behaves badly:
                # it'll get automatically encoded to some random encoding
                # you don't get to choose. All data should be encoded
                # manually by the user before it's passed to requests.
                encoded_chunk = chunk.encode("utf-8")
                response = requests.post(
                    channel_url, data=encoded_chunk, timeout=7
                )
                response.raise_for_status()
        except Exception as error:
            # Catch all network and HTTP errors raised by `requests`
            self.logger.error(
                "Slack request failed: {}".format(format_exception(error))
            )
